import logging, asyncio, sys, os
from pathlib import Path
from rich.logging import RichHandler
import concurrent.futures

# log=logging

class Utils:
  @staticmethod
  def log_handler(debug, msg, level: str):
    """Logging. All logging happens from here.

    Args:
        debug (bool): Should you log anything, or surpress warnings to
        msg (str/error): Message
        level (str): Choose type: (debug, info, warning, error, critical)
    """    
    if debug:
        FORMAT = "%(message)s"
        logging.basicConfig(
            level="NOTSET", format=FORMAT, datefmt="[%X]", handlers=[RichHandler()]
        )
        log = logging.getLogger("rich")
        if level.lower() == "debug":
            log.debug(msg)
        elif level.lower() == "info":
            log.info(msg)
        elif level.lower() == "warning":
            log.warning(msg)
        elif level.lower() == "error":
            log.error(msg)
        elif level.lower() == "critical":
            log.critical(msg)

  @staticmethod
  def run_async(function, *args):
    loop = asyncio.get_event_loop()
    out=loop.run_until_complete(function(*args))
    loop.close()
    return out
    
  # Disable
  @staticmethod
  def blockPrint():
    sys.stdout = open(os.devnull, 'w')

  # Restore
  @staticmethod
  def enablePrint():
    sys.stdout = sys.__stdout__
    
  @staticmethod
  def get_cwd():
    return os.getcwd()

  @staticmethod
  def check_file_exists(filepath: str, rest):
    # print("lkj")
    # print(path, filepath)
    # print(Path(path+"/"+filepath).is_file())
    return [Path(filepath).is_file(), rest]


from __future__ import unicode_literals
from lxml.cssselect import CSSSelector
from lxml.html import fromstring
import questionary
from rich import print, pretty
from rich.traceback import install
install()
pretty.install()
from rich.progress import track
from rich.console import Console
from rich.prompt import Prompt
console = Console()


def asker():
    try:
        single_songs = questionary.select(
            "How many songs do you want to download?",
            choices=[
                "One song",
                "Multiple Songs"
            ]
        ).ask()

        if single_songs == "One song":
            return [questionary.text("What is the song name").ask()]
        else:
            input_method = questionary.select(
                "Which songs do you want to download?",
                choices=[
                    "Songs, comma seperated",
                    'List in textfile (1 Song per line)',
                    'From YTMusic (beta)',
                ]).ask()  # returns value of selection

            if input_method == "From YTMusic":
                console.print("""[bold red]YT-Music[/bold red]. 
                - Go to your YTMusic liked songs playlist (https://music.youtube.com/playlist?list=LM)
                - Make sure you are logged in
                - Press Ctrl/Cmd + Shift + i and open the dev tools
                - Keep scrolling down until you reach the end of your playlist (Songs will stop loading)
                - Copy the all the html markup
                - Create a text file and paste the html into it.
                """)
                if not questionary.confirm("Only continue if you have done the task.").ask():
                    quit()
                file = questionary.path("Where is that file located?").ask()
                with open(file, "r", encoding="utf-8") as f:
                    data = f.read()
                h = fromstring(data)
                sel = CSSSelector("yt-formatted-string.title.style-scope.ytmusic-responsive-list-item-renderer.complex-string > a.yt-simple-endpoint.style-scope.yt-formatted-string")
                songs_list=[e.text for e in sel(h)]
                return songs_list

            elif input_method=="List in textfile (1 Song per line)":
                file = questionary.path("Where is that file located?").ask()
                text_file = open(file ,encoding='utf-8')
                songs_list = text_file.read().splitlines()
                return songs_list
            elif input_method=="Songs, comma seperated":
                songs_list = questionary.text("Write all songs (comma seperated)").ask().split(",")
                return songs_list
    except Exception as e:
        ## eventually change :/
        pass


###############################################################
# Published on ... under the MIT Licence
# Copyright 2021 techboy-coder
#                           _ _ 
#                         | | |
# _ __ ___  _ __ ___   __| | |
# | '_ ` _ \| '_ ` _ \ / _` | |
# | | | | | | | | | | | (_| | |   by techboy-coder
# |_| |_| |_|_| |_| |_|\__,_|_|   find me on https://github.com/techboy-coder
#                                                    
# mmdl [Mega Music Downloader] - A tool to easily download music.
###############################################################
from __future__ import unicode_literals
from lxml.cssselect import CSSSelector
from lxml.html import fromstring
import click
from mmdl import MusicDownloader
from .ask import asker
from rich.console import Console
import questionary
console = Console()
from click_help_colors import HelpColorsGroup, HelpColorsCommand
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

@click.group(context_settings=CONTEXT_SETTINGS, cls=HelpColorsGroup, help_headers_color='red', help_options_color='green')
def main():
  """\b                                             
                            _ _ 
                          | | |
  _ __ ___  _ __ ___   __| | |
  | '_ ` _ \| '_ ` _ \ / _` | |
  | | | | | | | | | | | (_| | |   by techboy-coder
  |_| |_| |_|_| |_| |_|\__,_|_|   find me on https://github.com/techboy-coder
                                                      
  mmdl [Mega Music Downloader] - A tool to easily download music.

  > Enter 'mmdl go' to start downloading songs.
  """
  pass

@main.command(context_settings=CONTEXT_SETTINGS, cls=HelpColorsCommand, help_headers_color='red', help_options_color='green')
@click.option("-v","--verbose", help="Verbose level", type=int, required=False, default=1)
@click.option("-d","--debug", help="Enable debug mode", is_flag=True)
def go(verbose, debug):
  """
  Easy mode. Just answer few questions and download your songs.

  mmdl [Music Downloader] - A tool to easily download music. \n
  
  # Go Command
  > Easy way to download songs. Uses inputs/inquirer. (Simply run 'mmdl go')
  """

  if debug:
    console.log("Verbose: ", verbose)
    console.log("Debug: ", debug)

  console.print("""\b[red]
                            _ _ 
                          | | |
  _ __ ___  _ __ ___   __| | |
  | '_ ` _ \| '_ ` _ \ / _` | |
  | | | | | | | | | | | (_| | |   [/]by techboy-coder[red]
  |_| |_| |_|_| |_| |_|\__,_|_|   [/]find me on https://github.com/techboy-coder
                                                      
  [green]mmdl [Mega Music Downloader] - A tool to easily download music.[/green]
  """)
  song_names=asker()
  console.print("\n[cyan]> [/] Total number of songs: %s. \n" % (len(song_names)))
  # List, verbose, debug
  MusicDownloader(song_names, verbose, debug).download_songs()

@main.group(context_settings=CONTEXT_SETTINGS, cls=HelpColorsGroup, help_headers_color='red', help_options_color='green')
def download():
  """
  Download your songs the way you want!

  mmdl [Mega Music Downloader] - A tool to easily download music. \n
  # Download Section
  > Here are possible ways to download songs.
  Ways to download:\n
  - Single song\n
  - Multiple songs\n
    - From file\n
    - As multiple songs (e.g. song1, song2, song3, ...)\n
    - From your YTMusic liked songs playlist (beta)\n
  """
  pass

@download.command(context_settings=CONTEXT_SETTINGS, cls=HelpColorsCommand, help_headers_color='red', help_options_color='green')
@click.argument("file", type=click.File("r"), required=True)
@click.option("-s","--seperator", help="Seperator of song_names", type=str, default="\n")
@click.option("-v","--verbose", help="Verbose level", type=int, required=False, default=1)
@click.option("-d","--debug", help="Enable debug mode", is_flag=True)
def file(file, seperator, verbose, debug):
  """
  Get songs from a file and then download them.

  mmdl [Music Downloader] - A tool to easily download music.

  ## File Section
  > Download multiple songs from a file.
  
  Run 'mmdl download file <filepath e.g. ~/Downloads/file.txt>'.
  """

  # print(type(file))
  # text_file = open(file ,encoding='utf-8')
  songs = file.read().split(seperator)
  if debug:
    console.log("File: ", file)
    console.log("Seperator: ", repr(seperator))
    console.log("Songs: ", str(songs))
    console.log("Verbose: ", verbose)
    console.log("Debug: ", debug)

  console.print("""\b[red]
                            _ _ 
                          | | |
  _ __ ___  _ __ ___   __| | |
  | '_ ` _ \| '_ ` _ \ / _` | |
  | | | | | | | | | | | (_| | |   [/]by techboy-coder[red]
  |_| |_| |_|_| |_| |_|\__,_|_|   [/]find me on https://github.com/techboy-coder
                                                      
  [green]mmdl [Mega Music Downloader] - A tool to easily download music.[/green]
  """)
  console.print("\n[cyan]> [/] Total number of songs: %s. \n" % (len(songs)))
  if not questionary.confirm("Do you want to continue").ask():
    quit()
  # file, verbose, debug
  MusicDownloader(songs, verbose, debug).download_songs()
  return
  
@download.command(context_settings=CONTEXT_SETTINGS, cls=HelpColorsCommand, help_headers_color='red', help_options_color='green')
@click.argument("songs", type=str, nargs=-1, required=False)
# @click.argument("seperator", type=str, required=True, nargs=1, default=",")
@click.option("-v","--verbose", help="Verbose level", type=int, required=False, default=1)
@click.option("-d","--debug", help="Enable debug mode", is_flag=True)
@click.option("-a","--ask", help="Get songs via input (easy)", is_flag=True)
def list(songs, verbose, debug, ask):
  """
  Download multiple songs.

  mmdl [Mega Music Downloader] - A tool to easily download music.

  ## List Section/Multiple Songs
  > Download multiple songs from a list.
  
  Run 'mmdl download list "Term1" "Term2" --ask (or -a)'.
  > Write all songs search terms (comma seperated): Term3, Term 4
  > This will download Term1, Term2, Term3, Term4

  """
  if ask:
    console.print("[cyan][>][/] We'll be manually asking you for songs.")
    songs_list = questionary.text("Write all songs search terms (comma seperated)").ask()
    if not songs_list:
      quit()
    songs_list = songs_list.split(",")
  else:
    if len(songs) < 1:
      console.print("[cyan][-][/] You didn't specify any songs. So we'll be manually asking them to you.")
      songs_list = questionary.text("Write all songs search terms (comma seperated)").ask()
      if not songs_list:
        quit()
      songs_list = songs_list.split(",")
    else:
      songs_list = []
    if isinstance(songs, tuple):
      s = [song for song in songs]+songs_list
      songs_list = s
  if debug:
    console.log("Songs: ", songs_list)
    console.log("Verbose: ", verbose)
    console.log("Debug: ", debug)

  console.print("""\b[red]
                            _ _ 
                          | | |
  _ __ ___  _ __ ___   __| | |
  | '_ ` _ \| '_ ` _ \ / _` | |
  | | | | | | | | | | | (_| | |   [/]by techboy-coder[red]
  |_| |_| |_|_| |_| |_|\__,_|_|   [/]find me on https://github.com/techboy-coder
                                                      
  [green]mmdl [Mega Music Downloader] - A tool to easily download music.[/green]
  """)
  console.print("\n[cyan]> [/] Total number of songs: %s. \n" % (len(songs_list)))
  # file, verbose, debug
  MusicDownloader(songs_list, verbose, debug).download_songs()
  return
  
@download.command(context_settings=CONTEXT_SETTINGS, cls=HelpColorsCommand, help_headers_color='red', help_options_color='green')
@click.argument("file", type=click.File("r"), required=False)
# @click.argument("seperator", type=str, required=True, nargs=1, default=",")
@click.option("-v","--verbose", help="Verbose level", type=int, required=False, default=1)
@click.option("-d","--debug", help="Enable debug mode", is_flag=True)
@click.option("-a","--ask", help="Get songs via input (easy)", is_flag=True)
def ytmusic(file, verbose, debug, ask):
  if not ask and not file:
    ask = True
  """
  Download multiple songs from YouTube Music Liked songs.

  mmdl [Mega Music Downloader] - A tool to easily download music.

  ## YTMusic Section
  > Download multiple songs from YTMusic liked songs (beta). [Done via parsing html.]
  
  Run 'mmdl download ytmusic <filepath e.g. ~/Downloads/file.txt> (or --ask or -a for entering file via input)'.

  """
  if ask:
    console.print("[cyan][>][/] We'll be manually asking you for the file location.")
    console.print("""
[bold red]YT-Music[/bold red]. 
- Go to your YTMusic liked songs playlist (https://music.youtube.com/playlist?list=LM)
- Make sure you are logged in
- Press Ctrl/Cmd + Shift + i and open the dev tools
- Keep scrolling down until you reach the end of your playlist (Songs will stop loading)
- Copy the all the html markup
- Create a text file and paste the html into it.
                """)
    if not questionary.confirm("Only continue if you have done the task.").ask():
        quit()
    file = questionary.path("Where is that file located?").ask()
    with open(file, "r", encoding="utf-8") as f:
        data = f.read()
    h = fromstring(data)
    sel = CSSSelector("yt-formatted-string.title.style-scope.ytmusic-responsive-list-item-renderer.complex-string > a.yt-simple-endpoint.style-scope.yt-formatted-string")
    songs_list=[e.text for e in sel(h)]
    if not songs_list[0]:
      console.log("[red][-] Hmm. No songs could be parsed from html. Did you select the correct HTML? If you see a error please make a bug report. Thanks!")
      quit()
  if not ask:
    if not file:
      console.log("[red][-] You need to enter the file location or add the -a flag.")
      quit()
    data = file.read()
    h = fromstring(data)
    sel = CSSSelector("yt-formatted-string.title.style-scope.ytmusic-responsive-list-item-renderer.complex-string > a.yt-simple-endpoint.style-scope.yt-formatted-string")
    songs_list=[e.text for e in sel(h)]
    if not songs_list[0]:
      console.log("[red][-] Hmm. No songs could be parsed from html. Did you select the correct HTML? If you see a error please make a bug report. Thanks!")
      console.print("""
[bold red]YT-Music[/bold red]. 
- Go to your YTMusic liked songs playlist (https://music.youtube.com/playlist?list=LM)
- Make sure you are logged in
- Press Ctrl/Cmd + Shift + i and open the dev tools
- Keep scrolling down until you reach the end of your playlist (Songs will stop loading)
- Copy the all the html markup
- Create a text file and paste the html into it.
      """)

  
  if debug:
    console.log("Songs: ", str(songs_list))
    console.log("Verbose: ", verbose)
    console.log("Debug: ", debug)
  console.print("""\b[red]
                            _ _ 
                          | | |
  _ __ ___  _ __ ___   __| | |
  | '_ ` _ \| '_ ` _ \ / _` | |
  | | | | | | | | | | | (_| | |   [/]by techboy-coder[red]
  |_| |_| |_|_| |_| |_|\__,_|_|   [/]find me on https://github.com/techboy-coder
                                                      
  [green]mmdl [Mega Music Downloader] - A tool to easily download music.[/green]
  """)
  console.print("\n[cyan]> [/] Total number of songs: %s. \n" % (len(songs_list)))
  if not questionary.confirm("Do you want to continue").ask():
    quit()
  # file, verbose, debug
  MusicDownloader(songs_list, verbose, debug).download_songs()
  return
  

@download.command(context_settings=CONTEXT_SETTINGS, cls=HelpColorsCommand, help_headers_color='red', help_options_color='green')
@click.argument("song", type=str, required=True, nargs=-1)
@click.option("-v","--verbose", help="Verbose level", type=int, required=False, default=1)
@click.option("-d","--debug", help="Enable debug mode", is_flag=True)
def single(song, verbose, debug):
  """
  Download a single song based on search query.

  mmdl [Mega Music Downloader] - A tool to easily download music.

  ## Single Song Section
  > Download a single song based on search query.
  
  Run 'mmdl download single <songname>'.
  """

  # print(type(file))
  song = " ".join(song)

  # text_file = open(file ,encoding='utf-8')
  if debug:
    console.log("File: ", song)
    console.log("Songs: ", str(songs))
    console.log("Verbose: ", verbose)
    console.log("Debug: ", debug)

  console.print("""\b[red]
                            _ _ 
                          | | |
  _ __ ___  _ __ ___   __| | |
  | '_ ` _ \| '_ ` _ \ / _` | |
  | | | | | | | | | | | (_| | |   [/]by techboy-coder[red]
  |_| |_| |_|_| |_| |_|\__,_|_|   [/]find me on https://github.com/techboy-coder
                                                      
  [green]mmdl [Mega Music Downloader] - A tool to easily download music.[/green]
  """)
  songs = [song]
  console.print("\n[cyan]> [/] Song: %s. \n" % (song))
  # file, verbose, debug
  MusicDownloader(songs, verbose, debug).download_songs()
  return
  

def runme():
  main()
if __name__ == '__main__':
  runme()
  # :/ 


from __future__ import unicode_literals
from lxml.cssselect import CSSSelector
from lxml.html import fromstring

import threading
import asyncio
import time

from pathlib import Path
import os, sys
# sys.path.append(os.getcwd())
import warnings
from functools import cache
import concurrent.futures

from youtubesearchpython.__future__ import VideosSearch
import youtube_dl
from ytmusicapi import YTMusic
import eyed3
import urllib

import logging
import questionary
from rich import print, pretty
from rich.traceback import install
install()
# pretty.install()
from rich.progress import track

from .globals import console

from .utils import Utils
from .yt_utils import YtdlUtils


##Dev Dep
import shutil

class DisplayablePath(object):
    display_filename_prefix_middle = '├──'
    display_filename_prefix_last = '└──'
    display_parent_prefix_middle = '    '
    display_parent_prefix_last = '│   '

    def __init__(self, path, parent_path, is_last):
        self.path = Path(str(path))
        self.parent = parent_path
        self.is_last = is_last
        if self.parent:
            self.depth = self.parent.depth + 1
        else:
            self.depth = 0

    @property
    def displayname(self):
        if self.path.is_dir():
            return self.path.name + '/'
        return self.path.name

    @classmethod
    def make_tree(cls, root, parent=None, is_last=False, criteria=None):
        root = Path(str(root))
        criteria = criteria or cls._default_criteria

        displayable_root = cls(root, parent, is_last)
        yield displayable_root

        children = sorted(list(path
                               for path in root.iterdir()
                               if criteria(path)),
                          key=lambda s: str(s).lower())
        count = 1
        for path in children:
            is_last = count == len(children)
            if path.is_dir():
                yield from cls.make_tree(path,
                                         parent=displayable_root,
                                         is_last=is_last,
                                         criteria=criteria)
            else:
                yield cls(path, displayable_root, is_last)
            count += 1

    @classmethod
    def _default_criteria(cls, path):
        return True

    @property
    def displayname(self):
        if self.path.is_dir():
            return self.path.name + '/'
        return self.path.name

    def displayable(self):
        if self.parent is None:
            return self.displayname

        _filename_prefix = (self.display_filename_prefix_last
                            if self.is_last
                            else self.display_filename_prefix_middle)

        parts = ['{!s} {!s}'.format(_filename_prefix,
                                    self.displayname)]

        parent = self.parent
        while parent and parent.parent is not None:
            parts.append(self.display_parent_prefix_middle
                         if parent.is_last
                         else self.display_parent_prefix_last)
            parent = parent.parent

        return ''.join(reversed(parts))

class MusicDownloader():
    def __init__(self, song_names: list[str], verbose: int, debug: bool):
        """Music Downloader Class. This is the main class :/

        Args:
            song_names (list[str]): list of all songs/titles you want to download.py
            verbose (int): Verbose level. 0 to INF. (1 and above=same)
            debug (bool): Enable debugging?
        """        
        
        self.song_names = song_names
        self.debug = debug
        self.verbose = verbose
        # self.output_path = download_dir_location
        self.ytdl = YtdlUtils(self)
    
    def find_multiple_songs_by_query(self):
        """Searches Youtube and finds songs by query. Uses "song_names" as a list of songs. Function itself is sync, but runs "async"-ly inside => speed.
        """        
        async def sub_finder(verbose, queries: list[str]):
            search_results = [asyncio.create_task(self.ytdl.find_song_on_yt(query)) for query in queries]
            search_results = await asyncio.gather(*search_results)
            results=[]
            if verbose >= 1:
                for task in track(
                    asyncio.as_completed(search_results),
                    description="Searching Youtube...",
                    total=len(search_results),
                ):
                    results.append(await task)
            else:
                for task in asyncio.as_completed(search_results):
                    results.append(await task)
            return results
        
        result=Utils.run_async(sub_finder, self.verbose, self.song_names)
        return result

    def download_songs(self):  # sourcery no-metrics
        """Downloads songs.
        """        
        start_time = time.time()
        songs_with_meta = self.find_multiple_songs_by_query()
        with concurrent.futures.ThreadPoolExecutor() as executor:
            temp = []
            if self.verbose >= 1:
                for song in songs_with_meta:
                    song_name = executor.submit(self.ytdl.download, song["result"][0]["link"], False, song)
                    temp.append(song_name)

                filenames = [
                    r.result()
                    for r in track(
                        concurrent.futures.as_completed(temp),
                        description="Getting file name(s)...",
                        total=len(temp),
                    )
                ]
                temp = []
                cwd = Utils.get_cwd()
                for f in filenames:
                    # print(f["filename"])
                    does_exist = executor.submit(Utils.check_file_exists, f["filename"], f)
                    temp.append(does_exist)
                exists = [
                    r.result()
                    for r in track(
                        concurrent.futures.as_completed(temp),
                        description="Checking for already existing file(s)...",
                        total=len(temp),
                    )
                ]
                temp=[]
                for f in exists:
                    if not f[0]:
                        song = executor.submit(self.ytdl.download, f[1]["url"], True, f[1])
                        temp.append(song)

                if len(temp) > 1:
                    for r in track(concurrent.futures.as_completed(temp), description="Downloading song(s) (Be patient, this can take some time)...", total=len(temp)):
                        pass
                else:
                    with console.status("[bold]Downloading single song, please wait...[/bold]", spinner="bouncingBar") as status:
                        for r in concurrent.futures.as_completed(temp):
                            pass

                temp=[]
                for f in exists:
                    song = executor.submit(self.ytdl.get_song_meta_by_v_id, f[1]["rest"]["result"][0]["id"], f)
                    temp.append(song)

                meta = [
                    r.result()
                    for r in track(
                        concurrent.futures.as_completed(temp),
                        description="Getting metadata for song(s)...",
                        total=len(temp),
                    )
                ]
                temp=[]
                for f in meta:
                    song = executor.submit(self.ytdl.change_metadata_of_song, f["rest"][1]["filename"],f["meta"]["videoDetails"]["author"],f["meta"]["videoDetails"]["title"],f["rest"][1]["rest"]["result"][0]["thumbnails"][0]["url"])
                    temp.append(song)
                meta = [
                    r.result()
                    for r in track(
                        concurrent.futures.as_completed(temp),
                        description="Writing metadata to song(s)...",
                        total=len(temp),
                    )
                ]
                if self.verbose > 1:
                    console.print("[cyan bold] \n\n\nAll songs ==> [/]")
                    for i in exists:
                        console.print("   • %s" % (i[1]["rest"]["result"][0]["title"]))
            else:
                console.log("[green][>][/green] Starting to download songs. (Verbose = Off)")
                for song in songs_with_meta:
                    song_name = executor.submit(self.ytdl.download, song["result"][0]["link"], False, song)
                    temp.append(song_name)

                filenames = [
                    r.result()
                    for r in concurrent.futures.as_completed(temp)
                ]
                temp = []
                cwd = Utils.get_cwd()
                for f in filenames:
                    does_exist = executor.submit(Utils.check_file_exists, f["filename"], f)
                    temp.append(does_exist)
                exists = [
                    r.result()
                    for r in concurrent.futures.as_completed(temp)
                ]
                temp=[]
                for f in exists:
                    if not f[0]:
                        song = executor.submit(self.ytdl.download, f[1]["url"], True, f[1])
                        temp.append(song)
                for r in concurrent.futures.as_completed(temp):
                    pass

                temp=[]
                for f in exists:
                    song = executor.submit(self.ytdl.get_song_meta_by_v_id, f[1]["rest"]["result"][0]["id"], f)
                    temp.append(song)

                meta = [
                    r.result()
                    for r in concurrent.futures.as_completed(temp)
                ]
                temp=[]
                for f in meta:
                    song = executor.submit(self.ytdl.change_metadata_of_song, f["rest"][1]["filename"],f["meta"]["videoDetails"]["author"],f["meta"]["videoDetails"]["title"],f["rest"][1]["rest"]["result"][0]["thumbnails"][0]["url"])
                    temp.append(song)
                meta = [
                    r.result()
                    for r in concurrent.futures.as_completed(temp)
                ]
            console.print("""\n[bold][green][+][/green] Successfully downloaded %s song(s). \n[green][+][/green] Total song(s): %s. \n[green][>][/green] Output directory: %s.  \n[green][>][/green] [cyan]Total time needed: %.2f second(s). [/cyan]
            [/bold]""" % (len(temp)-len([e for e in exists if e[0]]), len(temp), str(self.ytdl.output_dir.absolute()), (time.time() - start_time)))






from __future__ import unicode_literals
from .utils import Utils
import youtube_dl
from youtubesearchpython.__future__ import VideosSearch
from ytmusicapi import YTMusic
YTMusic = YTMusic()
from pathlib import Path
import os
import eyed3
import urllib
from .globals import console


class YtdlUtils():
    def __init__(self, mmdl: bool):

        self.debug = mmdl.debug
        self.verbose = mmdl.verbose
        self.output_dir = Path(Utils.get_cwd()+"/music")
        self.ytdl_opts = {
            'format': 'bestaudio/best',
            'postprocessors': [
                {
                'key': 'FFmpegExtractAudio',
                'preferredcodec': 'mp3',
                'preferredquality': '192',
                }
            ],
            'outtmpl': str(self.output_dir)+'/%(title)s.%(ext)s',
            'quiet': True,
            'ignoreerrors': True,
            "logger":self.logger(),
            'progress_hooks': [self.progress_hook()],
            'restrictfilenames':True,
            'forcefilename':True,
        }

    def download(self, url: str, do_download: bool, rest: None):
        if do_download:
            with youtube_dl.YoutubeDL(self.ytdl_opts) as ydl:
                info = ydl.extract_info(url, download=True)
                filename = ydl.prepare_filename(info)
                filename = filename.replace("webm","mp3")
                filename = filename.replace("m4a", "mp3")
                return {"filename": filename, "url": url, "rest":rest}
        else:
            with youtube_dl.YoutubeDL(self.ytdl_opts) as ydl:
                info = ydl.extract_info(url, download=False)
                filename = ydl.prepare_filename(info)
                filename = filename.replace("webm","mp3")
                filename = filename.replace("m4a", "mp3")
                return {"filename": filename, "url": url, "rest":rest}
            
    def logger(self):
        debug = self.debug
        class Logger(object):
            def debug(self, msg):
                Utils.log_handler(debug, msg, "debug")
            def warning(self, msg):
                Utils.log_handler(debug, msg, "warning")

            def error(self, msg):
                Utils.log_handler(debug, msg, "error")

        return Logger()
    
    def progress_hook(self):
        verbose = self.verbose
        def hook(d):

            pass

        return hook
    def clear_cache(self):
        with youtube_dl.YoutubeDL(self.ydl_opts) as ydl:
            try:
                if self.debug:
                    Utils.blockPrint()
                    ydl.cache.remove()
                    Utils.enablePrint()
                else:
                    ydl.cache.remove()
            except Exception as e:
                Utils.log_handler(self.debug, e, "error")
    
    async def find_song_on_yt(self, title: str):
        return VideosSearch(title, limit = 2).next()
    
    def get_song_meta_by_v_id(self, id, rest: None):
        return {"meta":YTMusic.get_song(id),"rest":rest}
    
    def change_metadata_of_song(self, filepath, artist, song_title, thumbnail_url):
        try:
            audiofile = eyed3.load(filepath)
            audiofile.tag.artist = artist
            audiofile.tag.title = song_title
            response = urllib.request.urlopen(thumbnail_url)  
            imagedata = response.read()
            audiofile.tag.images.set(3, imagedata , "image/jpeg" ,"Description")
            audiofile.tag.save()
        except Exception as e:
            Utils.log_handler(self.debug, e, "error")
        return filepath



from setuptools import setup, find_packages
import codecs
import os
with open('requirements.txt') as f:
    required = f.read().splitlines()

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

VERSION = '0.0.11'
DESCRIPTION = 'MMDL [Mega Music Downloader] - A tool to easily download music.'
LONG_DESCRIPTION = 'Music Downloading Cli Tool. Downloads audio and metadata from YouTube.'

# Setting up
setup(
    name="mmdl",
    version=VERSION,
    author="techboy-coder",
    # author_email="",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    url="https://github.com/techboy-coder/mmdl",
    packages=["mmdl"],
    install_requires=required,
    include_package_data=True,
    keywords=['python', 'audio', 'download', 'youtube', 'cli', 'tool'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.9",
    ],
    entry_points = {
        "console_scripts":[
            'mmdl=mmdl.cli:runme',
        ]
    }
)